# -*- coding: utf-8 -*-

import sys
import pathlib

sys.path.insert(0, str(pathlib.Path(__file__).parent.absolute())+"/../")
sys.path.insert(0, str(pathlib.Path(__file__).parent.absolute()))


from os.path import dirname, abspath
import SofaRuntime
from splib3.animation import AnimationManagerController

from sofagym.envs.Cath.CathToolbox import RewardShaper, GoalSetter

# Register all the common component in the factory.
SofaRuntime.importPlugin("Sofa.Component")

path = dirname(abspath(__file__)) + '/'


def add_plugins(root):
	root.addObject('RequiredPlugin', pluginName='BeamAdapter')
	root.addObject('RequiredPlugin', name='SofaOpenglVisual')

	root.addObject('RequiredPlugin', name='SofaBoundaryCondition')
	root.addObject('RequiredPlugin', name='SofaConstraint')
	root.addObject('RequiredPlugin', name='SofaDeformable')
	root.addObject('RequiredPlugin', name='SofaGeneralLinearSolver')
	root.addObject('RequiredPlugin', name='SofaImplicitOdeSolver')
	root.addObject('RequiredPlugin', name='SofaLoader')
	root.addObject('RequiredPlugin', name='SofaMeshCollision')
	root.addObject('RequiredPlugin', name='SofaTopologyMapping')

	return root


def add_visuals_and_solvers(root, config, mode_simu=True, mode_visu=True):
	if mode_visu:
		source = config["source"]
		target = config["target"]

		root.addObject('VisualStyle', displayFlags='showVisualModels hideBehaviorModels hideCollisionModels '
												   'hideMappings hideForceFields hideWireframe')
		# root.addObject('OglSceneFrame', style="Arrows", alignment="TopRight")

		root.addObject("LightManager")
		root.addObject("SpotLight", position=source, direction=[target[i] - source[i] for i in range(len(source))])
		root.addObject('InteractiveCamera', name='camera', position=source, lookAt=target, zFar=5000)
		# root.addObject('BackgroundSetting', color=[1, 1, 1, 1])
	root.addObject('FreeMotionAnimationLoop')
	lcp_solver = root.addObject('LCPConstraintSolver', mu=0.1, tolerance=3e-3, maxIt=1000, build_lcp=False)

	if mode_simu:
		root.addObject('DefaultPipeline', draw=False, depth=6, verbose=False)
		root.addObject('BruteForceBroadPhase')
		root.addObject('BVHNarrowPhase')

		root.addObject('GenericConstraintSolver', maxIterations=500, tolerance=1e-8)
		root.addObject('BruteForceDetection', name="N2")
		root.addObject('DefaultContactManager', name='Response', response="FrictionContactConstraint")
		root.addObject('LocalMinDistance', contactDistance='1', alarmDistance='5', name='localmindistance', angleCone='0.02')

		root.addObject(AnimationManagerController(root))

	return root, lcp_solver


def add_goal_node(root):
	goal = root.addChild("Goal")
	goal.addObject('VisualStyle', displayFlags="showCollisionModels")
	goal_mo = goal.addObject('MechanicalObject', name='GoalMO', showObject=True, drawMode="1", showObjectScale=2.0,
							 showColor=[0, 1, 1, 0.5], position=[0.0, 0.0, 0.0])
	return goal_mo


def add_instrument_topology(root, name, i):
	# Topo Guide
	topology_node = root.addChild('topoLines_'+name)
	topology_node.addObject('WireRestShape', name=name+'RestShape', straightLength=950, length=1000, numEdges=200,
							youngModulus=1e3, spireDiameter=70, numEdgesCollis=[50, 10], printLog=False,
							template='Rigid3d', spireHeight=0.0, densityOfBeams=[45, 7],
							youngModulusExtremity=1e3)
	topology_node.addObject('MechanicalObject', name='dofTopo2', template='Rigid3d')
	topology_node.addObject('EdgeSetTopologyContainer', name='meshLines'+name)
	topology_node.addObject('EdgeSetTopologyModifier', name='Modifier')
	topology_node.addObject('EdgeSetGeometryAlgorithms', name='GeomAlgo', template='Rigid3d')


def add_instruments_combined(root, instrument_list, mode_simu=True):
	# Instruments
	instrument_combined_node = root.addChild('InstrumentCombined')
	instrument_combined_node.addObject('VisualStyle', displayFlags='hideWireframe')
	# if mode_simu:
	instrument_combined_node.addObject('EulerImplicitSolver', rayleighStiffness=0.1, printLog=False, rayleighMass=0.1)
	instrument_combined_node.addObject('BTDLinearSolver', verification=False, subpartSolve=False, verbose=False)
	instrument_combined_node.addObject('RegularGridTopology', name='MeshLines', 
									nx=60, ny=1, nz=1,
									xmax=1.0, xmin=0.0, ymin=0, ymax=0, zmax=1, zmin=1,)
	instrument_combined_node.addObject('MechanicalObject', showIndices=False, name='DOFs', template='Rigid3d', ry=-90)

	for instrument in instrument_list:
		instrument_combined_node.addObject('WireBeamInterpolation',
										   WireRestShape='@../topoLines_'+instrument+'/'+instrument+'RestShape',
										   radius=1.15, printLog=False, name='Interpol'+instrument,
										   edgeList=[k for k in range(60)])
		instrument_combined_node.addObject('AdaptiveBeamForceFieldAndMass', massDensity=0.00000155,
										   name=instrument+'ForceField', interpolation='@Interpol'+instrument)

	instrument_combined_node.addObject('InterventionalRadiologyController', xtip=[1, 0, 0], name='m_ircontroller',
									   instruments=['Interpol'+instrument for instrument in instrument_list], step=5,
									   printLog=False, listening=True, template='Rigid3d',
									   startingPos=[0, 0, 0, 0, 0, 0, 1], rotationInstrument=[0, 0, 0], speed=0,
									   controlledInstrument=0)
	if mode_simu:
		instrument_combined_node.addObject('LinearSolverConstraintCorrection', wire_optimization='true', printLog=False)
	instrument_combined_node.addObject('FixedConstraint', indices=0, name='FixedConstraint')
	instrument_combined_node.addObject('RestShapeSpringsForceField', points='@m_ircontroller.indexFirstNode',
									   angularStiffness=1e8, stiffness=1e10)

	if mode_simu:
		# Collision model
		Collis = instrument_combined_node.addChild('Collis')
		Collis.activated = True
		Collis.addObject('EdgeSetTopologyContainer', name='collisEdgeSet')
		Collis.addObject('EdgeSetTopologyModifier', name='colliseEdgeModifier')
		Collis.addObject('MechanicalObject', name='CollisionDOFs')
		Collis.addObject('MultiAdaptiveBeamMapping', controller='../m_ircontroller',
						 useCurvAbs=True, printLog=False, name='collisMap')
		Collis.addObject('LineCollisionModel', proximity=0.0, group=1)
		Collis.addObject('PointCollisionModel', proximity=0.0, group=1)

	return instrument_combined_node


def add_instrument_visual(instrument_combined_node, instrument_name, color, i):
	visu_node = instrument_combined_node.addChild('Visu_'+instrument_name)
	visu_node.addObject('MechanicalObject', name='Quads')
	visu_node.addObject('QuadSetTopologyContainer', name='Container'+instrument_name)
	visu_node.addObject('QuadSetTopologyModifier', name='Modifier')
	visu_node.addObject('QuadSetGeometryAlgorithms', name='GeomAlgo', template='Vec3d')
	visu_node.addObject('Edge2QuadTopologicalMapping', radius=1.15, listening='true',
						input='@../../topoLines_'+instrument_name+'/meshLines'+instrument_name,
						nbPointsOnEachCircle='10', flipNormals='true', output='@Container'+instrument_name)
	visu_node.addObject('AdaptiveBeamMapping', interpolation='@../Interpol'+instrument_name,
						name='visuMap'+instrument_name, output='@Quads', isMechanical=False,
						input='@../DOFs', useCurvAbs=True, printLog=False)

	# Ogl model
	visu_ogl = visu_node.addChild('VisuOgl')
	visu_ogl.addObject('OglModel', color=color, quads='@../Container'+instrument_name+'.quads', name='Visual')
	visu_ogl.addObject('IdentityMapping', input='@../Quads', output='@Visual')


def add_geometry(root, config, mode_simu=True, mode_visu=True):
	Vessel = root.addChild('Vessel')
	Vessel.addObject('EulerImplicitSolver', name="ImplicitEulerSolver", rayleighMass=0.1, rayleighStiffness=0.1)
	Vessel.addObject('CGLinearSolver', name="CGLinearSolver", iterations=25, tolerance=1e-9, threshold=1e-9)
	Vessel.addObject('SparseGridTopology', fileTopology="/home/mark/Documents/SOFA/PA_catheterization_simulation/mesh/boya/7.final.obj", n="20 6 6")
	mo = Vessel.addObject('MechanicalObject', name="VesselState", scale=1, translation=[-20,-50, 100], rotation=[90 ,180 ,180])
	Vessel.addObject('UniformMass', totalMass=0.05)
	Vessel.addObject('HexahedronFEMForceField', template="Vec3d", name="FEM" ,poissonRatio=0.4, youngModulus=1000)
	Vessel.addObject('FixedConstraint', indices=[0, 1, 2, 3, 4, 7, 8])

	CollisionModel = Vessel.addChild('CollisionModel')
	CollisionModel.addObject('MeshSTLLoader', name="loader", filename='/home/mark/Documents/SOFA/PA_catheterization_simulation/mesh/boya/7.final.stl')

	CollisionModel.addObject('TriangleSetTopologyContainer',name="Container" ,src="@loader")
	CollisionModel.addObject('TriangleSetTopologyModifier',name="Modifier")
	#CollisionModel.addObject('TriangleSetTopologyAlgorithms',name="TopoAlgo",template="Vec3d")
	CollisionModel.addObject('TriangleSetGeometryAlgorithms',name="GeomAlgo",template="Vec3d")
	CollisionModel.addObject('MechanicalObject',name="CollisionState",src="@loader")
	CollisionModel.addObject('TriangleCollisionModel', template="Vec3d", name="Triangles", contactStiffness=1, bothSide=1)
	CollisionModel.addObject('LineCollisionModel', template="Vec3d", name="Lines", contactStiffness="@[-1].contactStiffness",bothSide=1)
	CollisionModel.addObject('PointCollisionModel', template="Vec3d", name="Points", contactStiffness="@[-1].contactStiffness",bothSide=1)
	CollisionModel.addObject('BarycentricMapping')

	VisualModel = CollisionModel.addChild('VisualModel')
	VisualModel.addObject('OglModel', name="Visual", material="Default Diffuse 1 0 0 1 0.3 Ambient 1 0 0 0.1 0.3",color=[0.8, 0.8, 0.8, 0.2])
	VisualModel.addObject('IdentityMapping', input="@..", output="@Visual" )

	return mo


DEFAULT_CONFIG = {"scene": "CTR",
				  "deterministic": True,
				  "source": [100, 260, 100],
				  "target": [0, 20, 100],
				  "mesh": "mesh/7.final.stl",
				  "scale": 1,
				  "rotation": [90, 180.0, 180],
				  "translation": [0, -90, 0],
				  "goalList": [[0.0, 0.0, 0.0]],
				  "scale_factor": 10,
				  "timer_limit": 50,
				  "timeout": 30,
				  "display_size": (800, 600),
				  "render": 1,
				  "save_data": False,
				  "save_path": path + "/Results" + "/CTR",
				  "planning": True,
				  "discrete": True,
				  "seed": 0,
				  "start_from_history": None,
				  "python_version": "python3.8",
				  "zFar": 5000,
				  }


def createScene(root, config=DEFAULT_CONFIG, tube_list=['tube_1'], mode='simu_and_visu'):

	print("SCENE ", config)

	# Choose the mode: visualization or computations (or both)
	visu, simu = False, False
	if 'visu' in mode:
		visu = True
	if 'simu' in mode:
		simu = True

	add_plugins(root)
	_, solver = add_visuals_and_solvers(root, config, simu, visu)
	root.dt = 0.01

	goal_mo = add_goal_node(root)
	for i, name in enumerate(tube_list):
		add_instrument_topology(root, name, i)

	instrument_combined_node = add_instruments_combined(root, tube_list, simu)
	color = [[0.0, 1.0, 0.0]]
	if visu:
		for i, name in enumerate(tube_list):
			add_instrument_visual(instrument_combined_node, name, color, i)

	cavity_mo = add_geometry(root, config, simu, visu)

	root.addObject(RewardShaper(name="Reward", rootNode=root, goalPos=config['goalPos']))
	root.addObject(GoalSetter(name="GoalSetter", rootNode=root, goalMO=goal_mo, goalPos=config['goalPos']))

	return root
